package cnf;

import static org.junit.Assert.*;

import org.junit.Test;
import cnf.*;
import sudoku.*;

public class _test_Solver {

	@Test
	public void test_empty_solution() {
		Formula f = new Formula();
		Environment e = new Environment();
		assertEquals(f.solve(e), null);
	}
	
	@Test
	public void test_empty_environment(){
		Formula f = new Formula();
		Environment e = new Environment();
		Clause c = new Clause();
		Literal l = new Literal(5);
		Literal l2 = new Literal(4);
		f.add(c);
		c.add(l);
		c.add(l2);
		
		Environment sol = f.solve(e);
		
		assertEquals(sol, null);	
	}
	
	@Test
	public void test_one_literal_solution(){
		Formula f = new Formula();
		Environment e = new Environment();
		Clause c = new Clause();
		Literal l = new Literal(5);
		f.add(c);
		c.add(l);
		e.add(5, Bool.UNDEFINED);
		
		Environment sol = f.solve(e);
		
		assertEquals(sol.eval(5), Bool.TRUE);
		assertEquals(sol.eval(0), Bool.UNDEFINED);
	}
	
	@Test
	/**
	 * Test the expression (a V b V c) ^ (~a V ~b V c)
	 */
	public void test_basic_satisfiable(){
		Formula f = new Formula();
		Environment e = new Environment();
		Clause one = new Clause(), two = new Clause();
		Literal a = new Literal(0), b = new Literal(1), c = new Literal(2);
		Literal _a = new Literal(0), _b = new Literal(1), _c = new Literal(2);
		_a.negate(); _b.negate(); //_c.negate();
		f.add(one); f.add(two);
		one.add(a); one.add(b); one.add(c);
		two.add(_a); two.add(_b); two.add(_c);
		e.initializeRange(3);
		
		Environment sol = f.solve(e);
		
		assertEquals(f.eval(sol), Bool.TRUE);
		assertEquals(sol.eval(2), Bool.TRUE);
		assertNotSame(sol.eval(1), Bool.UNDEFINED);
		assertNotSame(sol.eval(0), Bool.UNDEFINED);
	}	
	
	@Test
	/**
	 * Test unit prop on (a V FALSE V FALSE) ^ (~d V FALSE V FALSE)
	 */
	public void test_unit_prop_basic(){
		Formula f = new Formula();
		Environment e = new Environment();
		Clause one = new Clause(), two = new Clause();
		Literal a = new Literal(0), b = new Literal(1), c = new Literal(2);
		Literal _a = new Literal(2), _b = new Literal(1), _d = new Literal(3);
		_d.negate();
		f.add(one); f.add(two);
		one.add(a); one.add(b); one.add(c);
		two.add(_a); two.add(_b); two.add(_d);
		e.initializeRange(4);
		e.add(1, Bool.FALSE);
		e.add(2, Bool.FALSE);
		
		Environment prop = f.unitPropagate(e);
		
		assertEquals(f.eval(prop), Bool.TRUE);
		assertEquals(prop.eval(0), Bool.TRUE);
		assertEquals(prop.eval(3), Bool.FALSE);
	}
	
	@Test
	/**
	 * Test solve on HUGE formula representing blank 9x9 Sudoku
	 */
	public void test_blank_9_by_9_sudoku(){
		int size = 9;
		int blocksize = 3; //MUST BE SQRT of 4
		Formula f = new Formula();
		Environment e = new Environment();
		//CELL DEFINEDNESS (AT LEAST 1 NUMBER PER CELL)
		for(int r=0; r<size; r++)
		{
			for(int c=0; c<size; c++)
			{
				Clause cell_d = new Clause();
				f.add(cell_d);
				for(int v=0; v<size; v++)
				{
					Literal l = new Literal(r*size*size+c*size+v);
					cell_d.add(l);	
				}
			}
		}
		//CELL UNIQUENESS (NO MORE THAN 1 NUMBER PER CELL)
		for(int r=0; r<size; r++)
		{
			for(int c=0; c<size; c++)
			{
				for(int v1=0; v1<size-1; v1++)
				{
					for(int v2=v1+1; v2<size; v2++)
					{
						Clause cell_u = new Clause();
						f.add(cell_u);
						Literal l = new Literal(r*size*size+c*size+v1);
						cell_u.add(l); 
						l.negate(); 
						Literal l2 = new Literal(r*size*size+c*size+v2);
						l2.negate();
						cell_u.add(l2);	
					}
				}
			}
		}
		//ROW UNIQUENESS (NO MORE THAN 1 OF EACH VALUE PER ROW)
		for(int r=0; r<size; r++)
		{
			for(int v=0; v<size; v++)
			{
				for(int c1=0; c1<size-1; c1++)
				{
					for(int c2=c1+1; c2<size; c2++)
					{
						Clause row_u = new Clause();
						f.add(row_u);
						Literal l = new Literal(r*size*size+c1*size+v);
						Literal l2 = new Literal(r*size*size+c2*size+v);
						l.negate(); l2.negate();
						row_u.add(l); row_u.add(l2);	
					}
				}
			}
		}
		//COL UNIQUENESS (NO MORE THAN 1 NUMBER PER COL)
		for(int c=0; c<size; c++)
		{
			for(int v=0; v<size; v++)
			{
				for(int r1=0; r1<size-1; r1++)
				{
					for(int r2=r1+1; r2<size; r2++)
					{
						Clause col_u = new Clause();
						f.add(col_u);
						Literal l = new Literal(r1*size*size+c*size+v);
						Literal l2 = new Literal(r2*size*size+c*size+v);
						l.negate(); l2.negate();
						col_u.add(l); col_u.add(l2);	
					}
				}
			}
		}
		//BLOCK UNIQUENESS (NO MORE THAN 1 NUMBER PER BLOCK)
		for(int v=0; v<size; v++)
		{
			//select block # from 1, 2, ... n
			for(int a=0; a<size; a++)
			{
				//create ~a V ~b for all pairs within block
				for(int b=0; b<size-1; b++)
				{
					for(int c=b+1; c<size; c++)
					{
						int rStart = (a/blocksize)*blocksize; //rows are found by dividing by sqrt(size)
						int cStart = (a%blocksize)*blocksize; //cols are found by getting the remainder
						int r1 = rStart + b/blocksize;
						int r2 = rStart + c/blocksize;
						int c1 = cStart + b%blocksize;
						int c2 = cStart + c%blocksize;
						Clause block_u = new Clause();
						f.add(block_u);
						Literal l = new Literal(r1*size*size+c1*size+v);
						Literal l2 = new Literal(r2*size*size+c2*size+v);
						l.negate(); l2.negate();
						block_u.add(l); block_u.add(l2);
					}
				}
			}
		}
		e.initializeRange(size*size*size);
		Environment sol = f.solve(e);

		SudokuMachine.display(sol, size);
		assertNotSame(sol, null);
	}
	
}
