package sudoku;
import cnf.*;

import java.lang.String;
import java.io.*;

/**
 * Machine for reading Sudoku patterns from files, interpreting them as CNF formulas,
 * solving them using the cnf formula solver, and displaying them
 * @author scottyoung
 *
 */
public class SudokuMachine {

	private Formula puzzle;
	private boolean constraintsSet = false;
	private int size = 0;
	private Environment assigned;
	private Environment solution;
	
	/**
	 * Constructs a new Sudoku solver from the given filename
	 * throws an error if the file is incorrectly formatted
	 * @param filename
	 */
	public SudokuMachine(String filename) throws RuntimeException
	{
		boolean readError = false;
		puzzle = new Formula();
		assigned = new Environment();
		
		try{
			BufferedReader in = new BufferedReader(new FileReader(filename));
			
			int i=0;
			do
			{
				String line = in.readLine();
				size = line.length();
				
				//Before any set numbers are placed, initialize the entire set of board variables
				if(i==0)
					assigned.initializeRange(size*size*size);
			
				for(int j=0; j<size; j++)
				{
					int code = (int)line.charAt(j);
					if(code>=(int)'1' && code<=(int)'9')
					{
						//For that square, assign all [i,j,v] FALSE for v != # and TRUE for v == #
						for(int v=0; v<size; v++)
						{
							Bool value = Bool.FALSE;
							if(v == code-(int)'1')
								value = Bool.TRUE;
							assigned.add(i*size*size + j*size + v, value);
						}
					}
				}
				i++;
			}while(i<size);
			
			
		}catch(IOException e){
			throw(new RuntimeException("File was improperly read!"));
		}

		if(size != 1 && size != 4 && size != 9)
			throw(new RuntimeException("Invalid puzzle size. Size must be 1, 4 or 9."));
	}
	
	/**
	 * Attempts to solve the given Sudoku puzzle, returns true if
	 * successful, false if no solution is possible.
	 */
	public boolean solve()
	{
		
		if(assigned!=null && puzzle != null && size < 10 && solution == null)
		{
			constraints();
			solution = puzzle.solve(assigned);
		}
		
		if(solution==null)
			return false;
		else
			return true;
	}
	
	/**
	 * Creates a formula, given the input size, which constrains the SAT solver
	 */
	public void constraints()
	{
		if(!constraintsSet)
		{
			int blocksize = 1;
			if(size == 4)
				blocksize = 2;
			else if(size == 9)
				blocksize = 3;
			//CELL DEFINEDNESS (AT LEAST 1 NUMBER PER CELL)
			for(int r=0; r<size; r++)
			{
				for(int c=0; c<size; c++)
				{
					Clause cell_d = new Clause();
					puzzle.add(cell_d);
					for(int v=0; v<size; v++)
					{
						Literal l = new Literal(r*size*size+c*size+v);
						cell_d.add(l);	
					}
				}
			}
			//CELL UNIQUENESS (NO MORE THAN 1 NUMBER PER CELL)
			for(int r=0; r<size; r++)
			{
				for(int c=0; c<size; c++)
				{
					for(int v1=0; v1<size-1; v1++)
					{
						for(int v2=v1+1; v2<size; v2++)
						{
							Clause cell_u = new Clause();
							puzzle.add(cell_u);
							Literal l = new Literal(r*size*size+c*size+v1);
							cell_u.add(l); 
							l.negate(); 
							Literal l2 = new Literal(r*size*size+c*size+v2);
							l2.negate();
							cell_u.add(l2);	
						}
					}
				}
			}
			//ROW UNIQUENESS (NO MORE THAN 1 OF EACH VALUE PER ROW)
			for(int r=0; r<size; r++)
			{
				for(int v=0; v<size; v++)
				{
					for(int c1=0; c1<size-1; c1++)
					{
						for(int c2=c1+1; c2<size; c2++)
						{
							Clause row_u = new Clause();
							puzzle.add(row_u);
							Literal l = new Literal(r*size*size+c1*size+v);
							Literal l2 = new Literal(r*size*size+c2*size+v);
							l.negate(); l2.negate();
							row_u.add(l); row_u.add(l2);	
						}
					}
				}
			}
			//COL UNIQUENESS (NO MORE THAN 1 NUMBER PER COL)
			for(int c=0; c<size; c++)
			{
				for(int v=0; v<size; v++)
				{
					for(int r1=0; r1<size-1; r1++)
					{
						for(int r2=r1+1; r2<size; r2++)
						{
							Clause col_u = new Clause();
							puzzle.add(col_u);
							Literal l = new Literal(r1*size*size+c*size+v);
							Literal l2 = new Literal(r2*size*size+c*size+v);
							l.negate(); l2.negate();
							col_u.add(l); col_u.add(l2);	
						}
					}
				}
			}
			//BLOCK UNIQUENESS (NO MORE THAN 1 NUMBER PER BLOCK)
			for(int v=0; v<size; v++)
			{
				//select block # from 1, 2, ... n
				for(int a=0; a<size; a++)
				{
					//create ~a V ~b for all pairs within block
					for(int b=0; b<size-1; b++)
					{
						for(int c=b+1; c<size; c++)
						{
							int rStart = (a/blocksize)*blocksize; //rows are found by dividing by sqrt(size)
							int cStart = (a%blocksize)*blocksize; //cols are found by getting the remainder
							int r1 = rStart + b/blocksize;
							int r2 = rStart + c/blocksize;
							int c1 = cStart + b%blocksize;
							int c2 = cStart + c%blocksize;
							Clause block_u = new Clause();
							puzzle.add(block_u);
							Literal l = new Literal(r1*size*size+c1*size+v);
							Literal l2 = new Literal(r2*size*size+c2*size+v);
							l.negate(); l2.negate();
							block_u.add(l); block_u.add(l2);
						}
					}
				}
			}
			constraintsSet=true;
		}
	}
	
	/**
	 * Displays the Sudoku puzzle solution if available, otherwise
	 * displays the assigned keys
	 */
	public void display()
	{
		Environment board = assigned;
		
		if(solution != null)
			board = solution;
		
		display(board, size);
	
	}
	
	/**
	 * Display a particular solution (doesn't check if solution is valid)
	 */
	static public void display(Environment board, int size)
	{
		boolean outputted = false;
		int blocksize = 1;
		if(size == 4)
			blocksize = 2;
		else if(size == 9)
			blocksize = 3;
		
		System.err.println();
		
		for(int r=0; r<size; r++)
		{
			if(r%blocksize == 0)
			{
				for(int j=0; j<size+blocksize+1; j++)
					System.err.print("_");
				System.err.print("\n");
			}
			
			for(int c=0; c<size; c++)
			{
				outputted = false;
			
				if(c%blocksize == 0)
					System.err.print("|");
				
				for(int v=0; v<size; v++)
				{
					if(board.eval(r*size*size+c*size+v) == Bool.TRUE)
					{
						System.err.print((v+1));
						outputted = true;
					}
				}
				if(!outputted)
					System.err.print(" ");
			}
			System.err.print("|\n");
		}

		for(int j=0; j<size+blocksize+1; j++)
			System.err.print("_");
		System.err.print("\n");		
		
	}
	
}
