package abc;
import abc.*;
import music.*;
import sound.*;

public class MIDIVisitor implements MusicVisitor{

	private SequencePlayer midi;
	private static int TICKS = 400;
	
	//Variables for passing data between visit calls
	private int timeOffset = 0, nth;
	private double length = 1.0;
	private Key theKey;
	
	public MIDIVisitor(Music m)
	{
		try
		{
			int effectiveTempo = (int)(m.tempo*m.length*4);
			midi = new SequencePlayer(effectiveTempo, TICKS);
			theKey = m.key;
			length = m.length;
			
			//Now establish the sequence from the music
			visit(m);
			
		}catch(Exception e){
			System.err.println("Error: "+e.getMessage());
		}
	}
	
	public void play()
	{
		try{
			midi.play();
		}catch(Exception e){
			System.err.println("Error: "+e.getMessage());
		}
	}
	
	@Override
	public void visit(Music m) {
		for(Voice v : m.voices)
		{
			timeOffset = 0; //Each voice starts at the beginning
			v.accept(this);
		}
	}

	@Override
	public void visit(Voice v) {
		
		for(Section s : v.sections)
		{
			nth = 1;
			s.accept(this);
			if(s.repeat)
			{
				nth = 2;
				s.accept(this);
			}
		}
		
	}

	@Override
	public void visit(Section s) {
		
		for(Bar b : s.bars)
		{
			if(b.playOn == 0 || b.playOn == nth)
				b.accept(this);
		}
		
	}

	@Override
	public void visit(Bar b) {
		
		for(Elem e : b.elems)
		{
			e.accept(this);
		}
		
	}

	@Override
	public void visit(Chord c) {
		int timeStart = timeOffset;
		int longestTime = 0;
		
		for(Note n : c.notes)
		{
			n.accept(this);
			if(timeOffset-timeStart > longestTime)
				longestTime = timeOffset;
			timeOffset = timeStart;
		}
		timeOffset = longestTime;
		
	}

	@Override
	public void visit(Tuplet t) {
		for(Note n : t.notes)
			n.accept(this);
	}

	@Override
	public void visit(Note n) {
		
		char p = (char)n.pitch.toneToNote()[0];
		int semitones = n.pitch.toneToNote()[1];
		Accidental key = theKey.accidentals.get(n.pitch);
		if(key == null) key = Accidental.NONE;
		//First apply any accidental shifts in pitch
		switch(n.accidental)
		{
		case DOUBLEFLAT: semitones -= 2; break;
		case DOUBLESHARP: semitones += 2; break;
		case FLAT: semitones--; break;
		case SHARP: semitones++; break;
		case NATURAL: key = Accidental.NONE; break;
		}
		//Then apply the key to the pitch
		switch(key)
		{
		case FLAT: semitones--; break;
		case SHARP: semitones++; break;
		case NONE: break;
		default: System.err.println("Key stored a non #/b value.");
		}
		//Calculate the duration of the note, in TICKS
		int duration = (int)(length*n.length*TICKS*4);
		//Now add the note itself to the sequenceplayer and increment the timeOffset
		midi.addNote(new sound.Pitch(p).transpose(semitones).toMidiFrequency(), timeOffset, timeOffset + duration);
		timeOffset += duration;
	}

	@Override
	public void visit(Rest r) {
		
		timeOffset += (int)(r.duration()*length*TICKS*4);
		
	}

}
