package abc;

import music.*;
import abc.*;
import java.io.*;
import java.util.ArrayList;
import java.lang.String;
import java.util.regex.*;

public class Reader {
	
	public Music music;
	private Voice currVoice;
	private Section currSection;
	private Bar currBar;
	
	/**
	 * Creates a new Reader given the filename
	 * @param filename
	 */
	public Reader(String filename)
	{
		ArrayList<String> header = new ArrayList<String>();
		ArrayList<String> composition = new ArrayList<String>();
		String line = null;
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(filename));
			
			boolean inHeader = true;
			
			do
			{
				line = in.readLine();
				if(line != null)
				{
					if(line.matches("^[A-Z]:.*") && inHeader) //If the line is of the form "X: ... " it is a header
						header.add(line);
					else
						composition.add(line);
					if(line.matches("K:.*"))
						inHeader = false;
				}
				
			}while(line != null);
			
			
			music = new Music();
			
			parseHeader(header);
			parseComposition(composition);
			
			
		}catch(IOException e){
			System.err.println("Error reading from file."+e.getMessage());
		}
	}
	
	/**
	 * Parses the header into the music file categories, also asserts
	 * that order is X:, T:, ... , K: and outputs a warning to stderr
	 * if this is not true, but will not fail.
	 * @param header
	 */
	private void parseHeader(ArrayList<String> header)
	{
		int dataNumber = 0, keyField = 0;
		
		
		for(String datum: header)
		{
			dataNumber++;
			switch(datum.charAt(0))
			{
			case('X'): 
				music.trackNo = Integer.parseInt(datum.substring(2).trim()); break;
			case('T'):
				music.title = datum.substring(2).trim(); break;
			case('C'):
				music.composer = datum.substring(2).trim(); break;
			case('L'):
				music.length = Fraction.parseFraction(datum.substring(2).trim()); break;
			case('M'): 
				music.meter = Fraction.parseFraction(datum.substring(2).trim()); break;
			case('Q'):
				music.tempo = Integer.parseInt(datum.substring(2).trim()); break;
			case('K'):
				music.key = new Key(datum.substring(2).trim()); keyField = dataNumber; break;
			case('V'):
				music.addVoice(datum.substring(2).trim()); break;
			default:
				System.err.println("Warning: Using unsupported header tag, "+datum);
			}
			if(dataNumber==1 && datum.charAt(0) != 'X')
				System.err.println("Warning: First tag must be track number.");
			else if(dataNumber==2 && datum.charAt(0) != 'T')
				System.err.println("Warning: Second tag must be music title.");
		}
		if(dataNumber != keyField)
			System.err.println("Warning: Key must be the last header tag.");
	}
	
	/**
	 * Takes an array of strings which form the body of the music composition
	 * Assumes parseHeader has already been applied
	 * @param composition
	 */
	private void parseComposition(ArrayList<String> composition)
	{
		currVoice = music.addVoice("__DEFAULT__"); //Used when voices are undeclared
		currSection = null;
		currBar = null;
		
		
		for(String musicLine : composition)
		{
			if(musicLine.matches("%.*"))
			{
				//IGNORE COMMENTS
			}else if(musicLine.matches("V:.*")){
				//If we get a mid-line voice comment, switch voices
				currVoice = music.addVoice(musicLine.substring(2).trim());
				currSection = currVoice.newSection();
			}else{
				//We have a music line, so process this
				
				if(currSection == null) //if the section is null, create a new one
					currSection = currVoice.newSection();
				
				parseMusicLine(musicLine);
			}
		}
	}
	
	/**
	 * Takes a single line of music and parses it into its consituent elements
	 * @param m
	 */
	private void parseMusicLine(String m)
	{
		String[] chunks = m.split("\\|"); //splits into chunks based on barline characters
		for(int i = 0; i < chunks.length; i++)
		{
			currBar = currSection.newBar();
			if(chunks[i].matches("\\[[0-9].*")) //the bar should only be played on the nth repetition
			{
				int offset = Integer.parseInt(chunks[i].substring(1, 2));
				currBar.playOn = offset;
			}
			//Parse the bar
			parseBar(chunks[i]);
			//Check if we have an ending section or repeat
			if(chunks[i].matches(".*:"))
			{
				currSection = currVoice.newSection();
				currSection.repeat = true;
			}
		}
	}
	
	/**
	 * Takes a single bar and parses the individual notes
	 * @param b
	 */
	private void parseBar(String bar)
	{
		//we will process the bar by stripping it of elems
		String b = bar;
		while(b.length() > 0)
		{
			char c = b.charAt(0);
			boolean removed = false;
			int length = -1;
			//Case One: Starts with '[' so we have a chord (remember to filter out [1 and [2)
			if(c == '[' && (b.length() > 2 && !b.substring(1, 2).matches("[0-9]")))
			{
				int end = b.indexOf(']');
				if(end < 2)
					System.err.println("Improper syntax in: "+bar);
				else
					parseChord(b.substring(1, end));
				if(b.length() > end+1)
					b = b.substring(end+1);
				else
					b = "";
				removed=true;
			}
			
			//Case Two: starts with '(' so we have a tuplet
			if(c == '(')
			{
				length = parseTuplet(b.substring(1));
				if(b.length() > length+2)
					b = b.substring(length+2);
				else
					b = "";
				removed = true;
			}
			
			//Case Three: matches note expression, parse as note
			length = noteMatcher(b);
			if(length > 0)
			{
				Elem e = parseNote(b.substring(0, length));
				if(e instanceof Note)
					currBar.newNote((Note)e);
				else if(e instanceof Rest)
					currBar.newRest((Rest)e);
				else
					System.err.println("Tried to create elem which wasn't Note or Rest!");
				if(b.length() > length)
					b = b.substring(length);
				else
					b = "";
				removed = true;
			}
				
			//If we couldn't match to any elem pattern, just strip that character and try again
			if(!removed)
			{
				if(b.length() > 1)
					b = b.substring(1);
				else
					b = "";
			}
		}
	}
	
	/**
	 * Parses a chord
	 * @param c
	 */
	public void parseChord(String chord)
	{
		Chord c = currBar.newChord();
		String s = chord;
		while(s.length() > 0)
		{
			int length = noteMatcher(s);
			if(length > 0)
			{
				c.addNote((Note)parseNote(s.substring(0, length)));
				if(s.length() > length)
					s = s.substring(length);
				else
					s = "";
			}else{
				s = s.substring(1);
			}
		}
	}
	
	/**
	 * Parses a tuplet, returns the length of the used characters
	 * @param tuplet
	 * @return
	 */
	public int parseTuplet(String tuplet)
	{
		int n = 0, totalLength = 0;
		String t = tuplet.substring(1);
		Tuplet tup = currBar.newTuplet();
		
		if(tuplet.matches("[0-9].*"))
			n = Integer.parseInt(tuplet.substring(0, 1));
		else
		{
			System.err.println("Error: All tuplets must be preceded by # of members, was given: "+tuplet);
			return(tuplet.length());
		}
		//Parse the n members
		for(int i = 0; i < n; i++)
		{
			int length = noteMatcher(t);
			if(length <= 0)
			{
				System.err.println("Error: Tuplet couldn't be processed.");
				return(tuplet.length());
			} else {
				totalLength += length;
				tup.addNote((Note)parseNote(t.substring(0, length)));
				t = t.substring(length);
			}
		}
		tup.resize();
		return(totalLength);
	}
	
	/**
	 * Parses a note, supplying accidental, pitch and length information
	 * Unlike other parse commands, this returns the given note so it
	 * can be correctly appended to either a bar, tuplet or chord
	 * @param note
	 */
	public Elem parseNote(String note)
	{
		Elem e = null;
		Accidental a = Accidental.NONE;
		char p = 'C';
		Pitch pitch;
		double duration = 1.0;
		boolean isRest = false;
		
		
		//First, look for accidentals:
		if(note.length()>1)
		{
			if(note.substring(0, 2).equals("__"))
			{
				a = Accidental.DOUBLEFLAT;
				note = note.substring(2);
			}
			else if(note.substring(0, 2).equals("^^"))
			{
				a = Accidental.DOUBLESHARP;
				note = note.substring(2);
			}
			else if(note.charAt(0) == '_')
			{
				a = Accidental.FLAT;
				note = note.substring(1);
			}
			else if(note.charAt(0) == '^')
			{
				a = Accidental.SHARP;
				note = note.substring(1);		
			}
			else if(note.charAt(0) == '=')
			{
				a = Accidental.NATURAL;
				note = note.substring(1);			
			}
		}
	
		//Now look for pitch char
		if(note.charAt(0) == 'z' || note.charAt(0) == 'Z')
			isRest = true;
		else if(note.substring(0,1).matches("[CDEFGABcdefgab]"))
			p = note.charAt(0);
		else
			System.err.println("Couldn't read pitch information for: "+note);
		
		note = note.substring(1);
		
		//Now look for octave offsets
		int octaveOffset = 0;
		while(!note.equals("") && (note.charAt(0) == ',' || note.charAt(0) == '\''))
		{
			if(note.charAt(0) == ',')
				octaveOffset--;
			else
				octaveOffset++;
			note = note.substring(1);
		}
		pitch = Pitch.make(p, octaveOffset);
		
		//Finally transcribe length information
		if(note.length() > 0)
			duration = Fraction.parseFraction(note);
		
		if(isRest)
			e = new Rest(duration);
		else
			e = new Note(pitch, duration, a);
		
		return(e);
	}
	
	/**
	 * Examines the first characters in a string to see if they match
	 * the note pattern. If they do, return the length of that initial
	 * pattern for further processing.
	 * @param note
	 * @return
	 */
	public static int noteMatcher(String note)
	{
		int largest = 0;
		//returns only the largest to prevent clipping post-note information
		for(int i=1; i<=note.length(); i++)
		{
			if(note.substring(0, i).matches("(_|^|=){0,2}[CDEFGABZcdefgabz],*'*[0-9]*/?[0-9]*"))
				largest = i;
		}
		return(largest);
	}
	
	public static void main(String[] args)
	{
		Reader r = new Reader("paddy.abc");
		MIDIVisitor midi = new MIDIVisitor(r.music);
		//r.music.print();
		r.music.printHeader();
		midi.play();
	}

}
