package controller;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import model.Model;

public class Controller implements Runnable{
	
	//Queue the controller uses to process incoming client requests
	BlockingQueue<Protocol> requests;

	List<ClientHandler> clients;
	
	Model theModel;
	
	//Controls the amount of messages visible a chatroom
	final static int MSG_BUFFER_SIZE = 30;
	
	/**
	 * Establishes the server connection
	 */
	public Controller(){
		requests = new ArrayBlockingQueue<Protocol>(100000);
		clients = new ArrayList<ClientHandler>();
		theModel = new Model();
	}

	/**
	 * Starts the controller which consumes (and waits for) requests
	 * made to it in the form of Protocol objects, and then modifies
	 * the model and pushes messages to any clients.
	 */
	@Override
	public void run() {
		
		while(true)
		{
			try{
				//Get the next command
				Protocol action = requests.take();	
				
				execute(action);
				
				
			}catch(InterruptedException e){
				System.err.println("Error taking command from queue: "+e.getMessage());
			}
		}
		
	}
	
	public BlockingQueue<Protocol> getRequestQueue()
	{
		return requests;
	}
	
	/**
	 * Adds a new client to the client list (initiated by Server when accepting connections)
	 */
	public void addClient(ClientHandler c)
	{
		String name = theModel.newUser();
		clients.add(c);
		c.alias = name;
		//Give them their starting name
		c.send(new Protocol(Command.ASSIGNED, name));
		//Send a roomlist just to the new client, but send a userlist to everyone (updating users)
		List<Tuple> rooms = theModel.getRooms();
		c.send(new Protocol(Command.ROOMLIST, rooms));
		sendUserList();
	}
	
	/**
	 * Removes a client, given an alias
	 */
	public boolean removeClient(String alias)
	{
		boolean success = false;
		for(ClientHandler c : clients)
		{
			if(c.alias == alias)
			{
				success = true;
				clients.remove(c);
				break;
			}
		}
		return(success);	
	}
	
	/**
	 * Sends the USERLIST to all users
	 */
	public void sendUserList()
	{
		List<String> users = theModel.getUsers();
		for(ClientHandler c : clients)
			c.send(new Protocol(users));		
	}
	
	/**
	 * Executes the provided protocol by querying the model as well as sending
	 * messages to any clients who need updated information
	 */
	public void execute(Protocol p)
	{
		boolean success = true;
		switch(p.command)
		{
		
		//Create a new chatroom, send the client a message that it has been invited,
		//Send a message to all clients the updated ROOMLIST
		case CREATE:
			String name = theModel.newRoom(p.nameField);
			success = theModel.joinRoom(p.source.alias, name);
			p.source.send(new Protocol(Command.INVITED, name));
			List<Tuple> rooms = theModel.getRooms();
			for(ClientHandler c : clients)
			{
				c.send(new Protocol(Command.ROOMLIST, rooms));
			}
			break;
		//Join an existing chatroom, send the client a message that it has been invited, send them the transcript
		//Send a message to all clients the updated ROOMLIST
		case JOIN:
			success = theModel.joinRoom(p.source.alias, p.nameField);
			p.source.send(new Protocol(Command.INVITED, p.nameField)); //Let them know invite was successful
			
			rooms = theModel.getRooms();
			String transcript = theModel.getTranscript(p.nameField);
			p.source.send(new Protocol(Command.TRANSCRIPT, p.nameField, transcript)); 
			
			for(ClientHandler c : clients)
				c.send(new Protocol(Command.ROOMLIST, rooms));		
			
			break;
			
		//Exit an existing chatroom, send the ROOMLIST to all clients
		case LEAVE:
			success = theModel.leaveRoom(p.source.alias, p.nameField);
			rooms = theModel.getRooms();
			for(ClientHandler c : clients)
				c.send(new Protocol(Command.ROOMLIST, rooms));			
			break;
			
		//Change the alias of the client, send back the ASSIGNED name to the client and
		//send a message to all clients the updated USERLIST
		case ALIAS:
			name = theModel.changeAlias(p.source.alias, p.nameField);
			p.source.alias = name;
			p.source.send(new Protocol(Command.ASSIGNED, name));
			sendUserList();
			break;
			
		//Add the message to the chatroom, from the client. Send the transcript to all
		//clients inside the room
		case MSG:
			success = theModel.writeMessage(p.source.alias, p.nameField, p.textField);
			List<String> users = theModel.getAttendees(p.nameField);
			transcript = theModel.getTranscript(p.nameField, MSG_BUFFER_SIZE);
			for(ClientHandler c : clients)
				if(users.contains(c.alias))
					c.send(new Protocol(Command.TRANSCRIPT, p.nameField, transcript));
			break;
			
		//When client threads die, they send a kill request to remove the user and
		//should update all the clients with the new USERLIST and ROOMLIST (because capacity may have changed)
		case KILL:
			success = theModel.killUser(p.nameField);
			sendUserList();
			rooms = theModel.getRooms();
			for(ClientHandler c : clients)
				c.send(new Protocol(Command.ROOMLIST, rooms));	
			//Remove the client from the list of clients
			success = success && removeClient(p.nameField);
			break;	
			
		//Interpret an empty transcript command as a request for the transcript	
		case TRANSCRIPT:
			transcript = theModel.getTranscript(p.nameField, MSG_BUFFER_SIZE);
			p.source.send(new Protocol(Command.TRANSCRIPT, p.nameField, transcript));
			break;
			
		//Interpret an empty roomlist command as a request for the roomlist	
		case ROOMLIST:
			rooms = theModel.getRooms();
			p.source.send(new Protocol(Command.ROOMLIST, rooms));
			break;
			
		//Warning message for all protocols which the server cannot execute
		default:
			System.err.println("Warning, server should not be receiving message: "+p.command.toString());
			break;
		}
		
		//If any of the model queries failed, send a warning
		if(!success)
			System.err.println("Warning, error attempting to execute command: "+p.command.toString());
	}
	
	

}
